@@ -1,4 +1,25 @@
 #!/bin/bash
+
+## %CopyrightBegin%
+##
+## SPDX-License-Identifier: Apache-2.0
+##
+## Copyright Ericsson AB 2024-2025. All Rights Reserved.
+##
+## Licensed under the Apache License, Version 2.0 (the "License");
+## you may not use this file except in compliance with the License.
+## You may obtain a copy of the License at
+##
+##     http://www.apache.org/licenses/LICENSE-2.0
+##
+## Unless required by applicable law or agreed to in writing, software
+## distributed under the License is distributed on an "AS IS" BASIS,
+## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+## See the License for the specific language governing permissions and
+## limitations under the License.
+##
+## %CopyrightEnd%
+
 ## restore-from-prebuilt.sh CACHE_SRC_DIR TARGET [ARCHIVE]
 ##
 ## This script attempts to restore as much as possible from a previous
@@ -80,6 +101,8 @@ for delete in $DELETED; do
         rm -r "${CACHE_DIR}/${delete}"
     elif [ -f "${CACHE_DIR}/${delete}" ]; then
         rm "${CACHE_DIR}/${delete}"
+    elif [ "$(basename ${delete})" = ".gitignore" ]; then
+        echo "Should not delete .gitignore files"
     else
         echo "Could not find $delete to delete"
         exit 1
