@@ -1,7 +1,9 @@
 /*
  * %CopyrightBegin%
  *
- * Copyright Ericsson AB 2021-2024. All Rights Reserved.
+ * SPDX-License-Identifier: Apache-2.0
+ *
+ * Copyright Ericsson AB 2021-2025. All Rights Reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -39,7 +41,7 @@ extern "C"
 
 static std::string getAtom(Eterm atom) {
     Atom *ap = atom_tab(atom_val(atom));
-    return std::string((char *)ap->name, ap->len);
+    return std::string((char *)erts_atom_get_name(ap), ap->len);
 }
 
 BeamAssemblerCommon::BeamAssemblerCommon(BaseAssembler &assembler_)
@@ -808,7 +810,7 @@ void beam_jit_bs_add_argument_error(Process *c_p, Eterm A, Eterm B) {
 
 Eterm beam_jit_bs_init_bits(Process *c_p,
                             Eterm *reg,
-                            ERL_BITS_DECLARE_STATEP,
+                            ErlBitsState *EBS,
                             Uint num_bits,
                             Uint alloc,
                             unsigned Live) {
@@ -818,7 +820,7 @@ Eterm beam_jit_bs_init_bits(Process *c_p,
         alloc += ERL_REFC_BITS_SIZE;
     }
 
-    erts_bin_offset = 0;
+    EBS->erts_bin_offset = 0;
 
     if (num_bits <= ERL_ONHEAP_BITS_LIMIT) {
         ErlHeapBits *hb;
@@ -830,7 +832,7 @@ Eterm beam_jit_bs_init_bits(Process *c_p,
         hb->thing_word = header_heap_bits(num_bits);
         hb->size = num_bits;
 
-        erts_current_bin = (byte *)hb->data;
+        EBS->erts_current_bin = (byte *)hb->data;
         return make_bitstring(hb);
     } else {
         const Uint num_bytes = NBYTES(num_bits);
@@ -839,13 +841,13 @@ Eterm beam_jit_bs_init_bits(Process *c_p,
         test_bin_vheap(c_p, reg, num_bytes / sizeof(Eterm), alloc, Live);
 
         new_binary = erts_bin_nrml_alloc(num_bytes);
-        erts_current_bin = (byte *)new_binary->orig_bytes;
+        EBS->erts_current_bin = (byte *)new_binary->orig_bytes;
 
         return erts_wrap_refc_bitstring(&MSO(c_p).first,
                                         &MSO(c_p).overhead,
                                         &HEAP_TOP(c_p),
                                         new_binary,
-                                        erts_current_bin,
+                                        EBS->erts_current_bin,
                                         0,
                                         num_bits);
     }
@@ -1033,6 +1035,22 @@ Eterm beam_jit_int128_to_big(Process *p, Uint sign, Uint low, Uint high) {
     return make_big(hp);
 }
 
+void beam_jit_bs_put_binary_all(ErlBitsState *EBS, Process *c_p, Eterm arg) {
+    Uint offset, size;
+    byte *base;
+
+    ERTS_GET_BITSTRING(arg, base, offset, size);
+
+    copy_binary_to_buffer(EBS->erts_current_bin,
+                          EBS->erts_bin_offset,
+                          base,
+                          offset,
+                          size);
+    EBS->erts_bin_offset += size;
+
+    BUMP_REDS(c_p, size / ERL_BITS_PER_REDUCTION);
+}
+
 ErtsMessage *beam_jit_decode_dist(Process *c_p, ErtsMessage *msgp) {
     if (!erts_proc_sig_decode_dist(c_p, ERTS_PROC_LOCK_MAIN, msgp, 0)) {
         /*
@@ -1235,7 +1253,11 @@ void beam_jit_timeout_locked(Process *c_p) {
 void beam_jit_return_to_trace(Process *c_p,
                               Eterm session_weak_id,
                               Eterm *frame) {
-    if (ERTS_IS_P_TRACED_FL(c_p, F_TRACE_RETURN_TO)) {
+    ErtsTracerRef *ref =
+            get_tracer_ref_from_weak_id(&c_p->common, session_weak_id);
+
+    if (!ERTS_IS_PROC_SENSITIVE(c_p) && ref &&
+        IS_SESSION_TRACED_FL(ref, F_TRACE_RETURN_TO)) {
         ErtsCodePtr return_to_address;
         Uint *cpp;
 
@@ -1257,7 +1279,7 @@ void beam_jit_return_to_trace(Process *c_p,
         }
 
         ERTS_UNREQ_PROC_MAIN_LOCK(c_p);
-        erts_trace_return_to(c_p, return_to_address, session_weak_id);
+        erts_trace_return_to(c_p, return_to_address, ref);
         ERTS_REQ_PROC_MAIN_LOCK(c_p);
     }
 }
@@ -1277,16 +1299,16 @@ Eterm beam_jit_build_argument_list(Process *c_p, const Eterm *regs, int arity) {
     return res;
 }
 
-Export *beam_jit_handle_unloaded_fun(Process *c_p,
-                                     Eterm *reg,
-                                     int arity,
-                                     Eterm fun_thing) {
-    ErtsCodeIndex code_ix = erts_active_code_ix();
+const Export *beam_jit_handle_unloaded_fun(Process *c_p,
+                                           Eterm *reg,
+                                           int arity,
+                                           Eterm fun_thing) {
+    const ErtsCodeIndex code_ix = erts_active_code_ix();
+    const ErlFunEntry *fe;
+    const Export *ep;
     Eterm module, args;
     ErlFunThing *funp;
-    ErlFunEntry *fe;
     Module *modp;
-    Export *ep;
 
     funp = (ErlFunThing *)fun_val(fun_thing);
     ASSERT(is_local_fun(funp));
