@@ -1,7 +1,9 @@
 /*
  * %CopyrightBegin%
  *
- * Copyright Ericsson AB 2020-2023. All Rights Reserved.
+ * SPDX-License-Identifier: Apache-2.0
+ *
+ * Copyright Ericsson AB 2020-2025. All Rights Reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -132,7 +134,7 @@ void BeamGlobalAssembler::emit_plus_body_shared() {
     a.mov(TMP_MEM2q, ARG3);
 
     a.mov(ARG1, c_p);
-    runtime_call<3>(erts_mixed_plus);
+    runtime_call<Eterm (*)(Process *, Eterm, Eterm), erts_mixed_plus>();
 
     emit_leave_runtime();
     emit_leave_frame();
@@ -160,7 +162,7 @@ void BeamGlobalAssembler::emit_plus_guard_shared() {
 
     a.mov(ARG1, c_p);
     /* ARG2 and ARG3 were set by the caller */
-    runtime_call<3>(erts_mixed_plus);
+    runtime_call<Eterm (*)(Process *, Eterm, Eterm), erts_mixed_plus>();
 
     emit_leave_runtime();
     emit_leave_frame();
@@ -174,32 +176,46 @@ void BeamModuleAssembler::emit_i_plus(const ArgSource &LHS,
                                       const ArgSource &RHS,
                                       const ArgLabel &Fail,
                                       const ArgRegister &Dst) {
+    bool is_rhs_literal = false;
+    Sint rhs_untagged;
+
     bool small_result = is_sum_small_if_args_are_small(LHS, RHS);
 
+    if (RHS.isSmall()) {
+        /* Must be signed for the template magic in isInt32 to work for
+         * negative numbers. */
+        rhs_untagged = RHS.as<ArgSmall>().getSigned() << _TAG_IMMED1_SIZE;
+        is_rhs_literal = Support::isInt32(rhs_untagged);
+    }
+
     if (always_small(LHS) && always_small(RHS) && small_result) {
-        /* Since we don't need the order on this path (no exceptions), we'll
-         * simplify the code below by shuffling constants to the right-hand
-         * side. */
-        const ArgSource A = LHS.isSmall() ? RHS : LHS,
-                        B = LHS.isSmall() ? LHS : RHS;
-
         comment("add without overflow check");
-        mov_arg(RET, A);
 
-        if (B.isSmall()) {
-            /* Must be signed for the template magic in isInt32 to work for
-             * negative numbers. */
-            Sint untagged = B.as<ArgSmall>().getSigned() << _TAG_IMMED1_SIZE;
+        mov_arg(RET, LHS);
 
-            if (Support::isInt32(untagged)) {
-                a.add(RET, imm(untagged));
-            } else {
-                mov_imm(ARG2, B.as<ArgSmall>().get() & ~_TAG_IMMED1_MASK);
-                a.add(RET, ARG2);
-            }
+        if (is_rhs_literal) {
+            preserve_cache(
+                    [&]() {
+                        a.add(RET, imm(rhs_untagged));
+                    },
+                    RET);
+        } else if (RHS.isSmall()) {
+            preserve_cache(
+                    [&]() {
+                        mov_imm(ARG2,
+                                RHS.as<ArgSmall>().get() & ~_TAG_IMMED1_MASK);
+                        a.add(RET, ARG2);
+                    },
+                    RET,
+                    ARG2);
         } else {
-            mov_arg(ARG2, B);
-            a.lea(RET, x86::qword_ptr(RET, ARG2, 0, -_TAG_IMMED1_SMALL));
+            mov_arg(ARG2, RHS);
+            preserve_cache(
+                    [&]() {
+                        a.lea(RET,
+                              x86::qword_ptr(RET, ARG2, 0, -_TAG_IMMED1_SMALL));
+                    },
+                    RET);
         }
 
         mov_arg(Dst, RET);
@@ -210,16 +226,21 @@ void BeamModuleAssembler::emit_i_plus(const ArgSource &LHS,
 
     mov_arg(ARG2, LHS); /* Used by erts_mixed_plus in this slot */
     mov_arg(ARG3, RHS); /* Used by erts_mixed_plus in this slot */
-    emit_are_both_small(mixed, LHS, ARG2, RHS, ARG3);
 
-    a.mov(RET, ARG2);
-    a.and_(RET, imm(~_TAG_IMMED1_MASK));
-    a.add(RET, ARG3);
-    if (small_result) {
-        comment("skipped overflow test because the result is always small");
-        a.short_().jmp(next);
+    if (LHS.isLiteral() || RHS.isLiteral()) {
+        comment("skipped test for small because one operand is never small");
     } else {
-        a.short_().jno(next);
+        emit_are_both_small(mixed, LHS, ARG2, RHS, ARG3);
+
+        if (small_result) {
+            comment("skipped overflow test because the result is always small");
+            a.lea(RET, x86::qword_ptr(ARG2, ARG3, 0, -_TAG_IMMED1_SMALL));
+            a.short_().jmp(next);
+        } else {
+            a.lea(RET, x86::qword_ptr(ARG2, -_TAG_IMMED1_SMALL));
+            a.add(RET, ARG3);
+            a.short_().jno(next);
+        }
     }
 
     /* Call mixed addition. */
@@ -250,7 +271,7 @@ void BeamGlobalAssembler::emit_minus_body_shared() {
     a.mov(TMP_MEM2q, ARG3);
 
     a.mov(ARG1, c_p);
-    runtime_call<3>(erts_mixed_minus);
+    runtime_call<Eterm (*)(Process *, Eterm, Eterm), erts_mixed_minus>();
 
     emit_leave_runtime();
     emit_leave_frame();
@@ -278,7 +299,7 @@ void BeamGlobalAssembler::emit_minus_guard_shared() {
 
     a.mov(ARG1, c_p);
     /* ARG2 and ARG3 were set by the caller */
-    runtime_call<3>(erts_mixed_minus);
+    runtime_call<Eterm (*)(Process *, Eterm, Eterm), erts_mixed_minus>();
 
     emit_leave_runtime();
     emit_leave_frame();
@@ -292,23 +313,30 @@ void BeamModuleAssembler::emit_i_minus(const ArgSource &LHS,
                                        const ArgSource &RHS,
                                        const ArgLabel &Fail,
                                        const ArgRegister &Dst) {
+    bool is_rhs_literal = false;
+    Sint rhs_untagged = 0;
     bool small_result = is_diff_small_if_args_are_small(LHS, RHS);
 
+    if (RHS.isSmall()) {
+        /* Must be signed for the template magic in isInt32 to work for
+         * negative numbers. */
+        rhs_untagged = RHS.as<ArgSmall>().getSigned() << _TAG_IMMED1_SIZE;
+        is_rhs_literal = Support::isInt32(rhs_untagged);
+    }
+
     if (always_small(LHS) && always_small(RHS) && small_result) {
         comment("subtract without overflow check");
         mov_arg(RET, LHS);
 
-        if (RHS.isSmall()) {
-            /* Must be signed for the template magic in isInt32 to work for
-             * negative numbers. */
-            Sint untagged = RHS.as<ArgSmall>().getSigned() << _TAG_IMMED1_SIZE;
-
-            if (Support::isInt32(untagged)) {
-                a.sub(RET, imm(untagged));
-            } else {
-                mov_imm(ARG2, RHS.as<ArgSmall>().get() & ~_TAG_IMMED1_MASK);
-                a.sub(RET, ARG2);
-            }
+        if (is_rhs_literal) {
+            preserve_cache(
+                    [&]() {
+                        a.sub(RET, imm(rhs_untagged));
+                    },
+                    RET);
+        } else if (RHS.isSmall()) {
+            mov_imm(ARG2, RHS.as<ArgSmall>().get() & ~_TAG_IMMED1_MASK);
+            a.sub(RET, ARG2);
         } else {
             mov_arg(ARG2, RHS);
             a.and_(ARG2, imm(~_TAG_IMMED1_MASK));
@@ -324,20 +352,31 @@ void BeamModuleAssembler::emit_i_minus(const ArgSource &LHS,
     mov_arg(ARG2, LHS); /* Used by erts_mixed_plus in this slot */
     mov_arg(ARG3, RHS); /* Used by erts_mixed_plus in this slot */
 
-    emit_are_both_small(mixed, LHS, ARG2, RHS, ARG3);
-
-    if (small_result) {
-        comment("skipped overflow test because the result is always small");
-        a.mov(RET, ARG2);
-        a.and_(ARG3, imm(~_TAG_IMMED1_MASK));
-        a.sub(RET, ARG3);
-        a.short_().jmp(next);
+    if (LHS.isLiteral() || RHS.isLiteral()) {
+        comment("skipped test for small because one operand is never small");
     } else {
-        a.mov(RET, ARG2);
-        a.mov(ARG4, ARG3);
-        a.and_(ARG4, imm(~_TAG_IMMED1_MASK));
-        a.sub(RET, ARG4);
-        a.short_().jno(next);
+        emit_are_both_small(mixed, LHS, ARG2, RHS, ARG3);
+
+        if (small_result) {
+            comment("skipped overflow test because the result is always small");
+            a.mov(RET, ARG2);
+            if (is_rhs_literal) {
+                a.sub(RET, imm(rhs_untagged));
+            } else {
+                a.and_(ARG3, imm(~_TAG_IMMED1_MASK));
+                a.sub(RET, ARG3);
+            }
+            a.short_().jmp(next);
+        } else {
+            a.mov(RET, ARG2);
+            if (is_rhs_literal) {
+                a.sub(RET, imm(rhs_untagged));
+            } else {
+                a.lea(ARG4, x86::qword_ptr(ARG3, -_TAG_IMMED1_SMALL));
+                a.sub(RET, ARG4);
+            }
+            a.short_().jno(next);
+        }
     }
 
     a.bind(mixed);
@@ -364,7 +403,7 @@ void BeamGlobalAssembler::emit_unary_minus_body_shared() {
     a.mov(TMP_MEM1q, ARG2);
 
     a.mov(ARG1, c_p);
-    runtime_call<2>(erts_unary_minus);
+    runtime_call<Eterm (*)(Process *, Eterm), erts_unary_minus>();
 
     emit_leave_runtime();
     emit_leave_frame();
@@ -390,7 +429,7 @@ void BeamGlobalAssembler::emit_unary_minus_guard_shared() {
 
     a.mov(ARG1, c_p);
     /* ARG2 was set by the caller */
-    runtime_call<2>(erts_unary_minus);
+    runtime_call<Eterm (*)(Process *, Eterm), erts_unary_minus>();
 
     emit_leave_runtime();
     emit_leave_frame();
@@ -403,7 +442,7 @@ void BeamGlobalAssembler::emit_unary_minus_guard_shared() {
 void BeamModuleAssembler::emit_i_unary_minus(const ArgSource &Src,
                                              const ArgLabel &Fail,
                                              const ArgRegister &Dst) {
-    ArgVal zero = ArgVal(ArgVal::Immediate, make_small(0));
+    ArgVal zero = ArgVal(ArgVal::Type::Immediate, make_small(0));
     bool small_result = is_diff_small_if_args_are_small(zero, Src);
 
     if (always_small(Src) && small_result) {
@@ -506,7 +545,8 @@ void BeamGlobalAssembler::emit_int_div_rem_guard_shared() {
         a.lea(ARG4, TMP_MEM1q);
         a.lea(ARG5, TMP_MEM2q);
         a.mov(ARG1, c_p);
-        runtime_call<5>(erts_int_div_rem);
+        runtime_call<int (*)(Process *, Eterm, Eterm, Eterm *, Eterm *),
+                     erts_int_div_rem>();
 
         emit_leave_runtime();
 
@@ -592,7 +632,8 @@ void BeamGlobalAssembler::emit_int_div_rem_body_shared() {
         a.lea(ARG4, TMP_MEM4q);
         a.lea(ARG5, TMP_MEM5q);
         a.mov(ARG1, c_p);
-        runtime_call<5>(erts_int_div_rem);
+        runtime_call<int (*)(Process *, Eterm, Eterm, Eterm *, Eterm *),
+                     erts_int_div_rem>();
 
         emit_leave_runtime();
         emit_leave_frame();
@@ -652,10 +693,10 @@ void BeamModuleAssembler::emit_div_rem(const ArgLabel &Fail,
         divisor = RHS.as<ArgSmall>().getSigned();
     }
 
-    if (divisor != (Sint)0 && divisor != (Sint)-1) {
+    mov_arg(x86::rax, LHS);
+
+    if (divisor != 0 && divisor != -1) {
         /* There is no possibility of overflow. */
-        a.mov(ARG6, imm(divisor));
-        mov_arg(x86::rax, LHS);
         if (always_small(LHS)) {
             comment("skipped test for small dividend since it is always small");
             need_generic = false;
@@ -672,10 +713,9 @@ void BeamModuleAssembler::emit_div_rem(const ArgLabel &Fail,
             a.short_().jne(generic_div);
         }
 
-        /* Sign-extend and divide. The result is implicitly placed in
-         * RAX and the remainder in RDX (ARG3). */
         if (Support::isPowerOf2(divisor) &&
             std::get<0>(getClampedRange(LHS)) >= 0) {
+            /* Unsigned integer division. */
             int trailing_bits = Support::ctz<Eterm>(divisor);
 
             if (need_rem) {
@@ -692,8 +732,52 @@ void BeamModuleAssembler::emit_div_rem(const ArgLabel &Fail,
                 a.shr(x86::rax, imm(trailing_bits));
                 a.or_(x86::rax, imm(_TAG_IMMED1_SMALL));
             }
+        } else if (Support::isPowerOf2(divisor)) {
+            /* Signed integer division. */
+            int shift = Support::ctz<Eterm>(divisor);
+            Sint offset = (divisor - 1) << _TAG_IMMED1_SIZE;
+
+            if (need_rem) {
+                a.mov(x86::rdx, x86::rax);
+                ASSERT(x86::rdx != ARG1);
+            }
+
+            if (need_div) {
+                comment("optimized div by replacing with right shift");
+            }
+
+            if (divisor == 2) {
+                ERTS_CT_ASSERT(_TAG_IMMED1_SMALL == _TAG_IMMED1_MASK);
+                a.mov(ARG1, x86::rax);
+                a.shr(ARG1, imm(63));
+                a.add(x86::rax, ARG1);
+            } else {
+                if (Support::isInt32(offset)) {
+                    a.lea(ARG1, x86::qword_ptr(x86::rax, offset));
+                } else {
+                    a.mov(ARG1, offset);
+                    a.add(ARG1, x86::rax);
+                }
+                a.test(x86::rax, x86::rax);
+                a.cmovs(x86::rax, ARG1);
+            }
+
+            if (need_rem) {
+                Uint mask = (Uint)-1 << (shift + _TAG_IMMED1_SIZE);
+                comment("optimized rem by replacing with subtraction");
+                mov_imm(ARG1, mask);
+                a.and_(ARG1, x86::rax);
+                a.sub(x86::rdx, ARG1);
+            }
+
+            if (need_div) {
+                ERTS_CT_ASSERT(_TAG_IMMED1_SMALL == _TAG_IMMED1_MASK);
+                a.sar(x86::rax, imm(shift));
+                a.or_(x86::rax, imm(_TAG_IMMED1_SMALL));
+            }
         } else {
             comment("divide with inlined code");
+            a.mov(ARG6, imm(divisor));
             a.sar(x86::rax, imm(_TAG_IMMED1_SIZE));
             a.cqo();
             a.idiv(ARG6);
@@ -723,7 +807,7 @@ void BeamModuleAssembler::emit_div_rem(const ArgLabel &Fail,
     a.bind(generic_div);
     if (need_generic) {
         mov_arg(ARG4, RHS); /* Done first as mov_arg may clobber ARG1 */
-        mov_arg(ARG1, LHS);
+        a.mov(ARG1, x86::rax);
 
         if (Fail.get() != 0) {
             safe_fragment_call(ga->get_int_div_rem_guard_shared());
@@ -800,7 +884,7 @@ void BeamModuleAssembler::emit_i_m_div(const ArgLabel &Fail,
 
     /* Must be set last since mov_arg() may clobber ARG1 */
     a.mov(ARG1, c_p);
-    runtime_call<3>(erts_mixed_div);
+    runtime_call<Eterm (*)(Process *, Eterm, Eterm), erts_mixed_div>();
 
     emit_leave_runtime();
 
@@ -823,16 +907,32 @@ void BeamModuleAssembler::emit_i_m_div(const ArgLabel &Fail,
     mov_arg(Dst, RET);
 }
 
-/* ARG2 = LHS, ARG3 (!) = RHS
+/* ARG2 = Src1
+ * ARG3 = Src2
+ * ARG4 = Increment
  *
  * Result is returned in RET, error is indicated by ZF. */
-void BeamGlobalAssembler::emit_times_guard_shared() {
+void BeamGlobalAssembler::emit_mul_add_guard_shared() {
+    Label done = a.newLabel();
+
     emit_enter_frame();
     emit_enter_runtime();
 
-    a.mov(ARG1, c_p);
-    runtime_call<3>(erts_mixed_times);
+    a.mov(TMP_MEM1q, ARG4);
 
+    a.mov(ARG1, c_p);
+    runtime_call<Eterm (*)(Process *, Eterm, Eterm), erts_mixed_times>();
+    emit_test_the_non_value(RET);
+    a.short_().je(done);
+
+    a.mov(ARG3, TMP_MEM1q);
+    a.mov(ARG2, RET);
+    a.mov(ARG1, c_p);
+    a.cmp(ARG3, imm(make_small(0)));
+    a.short_().je(done);
+    runtime_call<Eterm (*)(Process *, Eterm, Eterm), erts_mixed_plus>();
+
+    a.bind(done);
     emit_leave_runtime();
     emit_leave_frame();
 
@@ -841,13 +941,14 @@ void BeamGlobalAssembler::emit_times_guard_shared() {
     a.ret();
 }
 
-/* ARG2 = LHS, ARG3 (!) = RHS
+/* ARG2 = Src1
+ * ARG3 = Src2
+ * ARG4 = Increment
  *
  * Result is returned in RET. */
-void BeamGlobalAssembler::emit_times_body_shared() {
-    static const ErtsCodeMFA bif_mfa = {am_erlang, am_Times, 2};
-
-    Label error = a.newLabel();
+void BeamGlobalAssembler::emit_mul_add_body_shared() {
+    Label mul_only = a.newLabel(), error = a.newLabel(),
+          mul_error = a.newLabel(), do_error = a.newLabel();
 
     emit_enter_frame();
     emit_enter_runtime();
@@ -855,61 +956,168 @@ void BeamGlobalAssembler::emit_times_body_shared() {
     /* Save original arguments for the error path. */
     a.mov(TMP_MEM1q, ARG2);
     a.mov(TMP_MEM2q, ARG3);
-
     a.mov(ARG1, c_p);
-    runtime_call<3>(erts_mixed_times);
+    a.cmp(ARG4, imm(make_small(0)));
+    a.short_().je(mul_only);
+    a.mov(TMP_MEM4q, ARG4);
+
+    a.lea(ARG5, TMP_MEM3q);
+    runtime_call<Eterm (*)(Process *, Eterm, Eterm, Eterm, Eterm *),
+                 erts_mul_add>();
 
     emit_leave_runtime();
     emit_leave_frame();
 
     emit_test_the_non_value(RET);
     a.short_().je(error);
+
     a.ret();
 
+    a.bind(mul_only);
+    {
+        runtime_call<Eterm (*)(Process *, Eterm, Eterm), erts_mixed_times>();
+
+        emit_leave_runtime();
+        emit_leave_frame();
+
+        emit_test_the_non_value(RET);
+        a.short_().je(mul_error);
+
+        a.ret();
+    }
+
     a.bind(error);
     {
-        /* Place the original arguments in x-registers. */
+        static const ErtsCodeMFA mul_mfa = {am_erlang, am_Times, 2};
+        static const ErtsCodeMFA add_mfa = {am_erlang, am_Plus, 2};
+
+        a.mov(ARG1, TMP_MEM3q);
+        a.mov(ARG2, TMP_MEM4q);
+        mov_imm(ARG4, &add_mfa);
+        emit_test_the_non_value(ARG1);
+        a.short_().jne(do_error);
+
+        a.bind(mul_error);
         a.mov(ARG1, TMP_MEM1q);
         a.mov(ARG2, TMP_MEM2q);
+        mov_imm(ARG4, &mul_mfa);
+
+        a.bind(do_error);
         a.mov(getXRef(0), ARG1);
         a.mov(getXRef(1), ARG2);
-
-        a.mov(ARG4, imm(&bif_mfa));
         a.jmp(labels[raise_exception]);
     }
 }
 
-void BeamModuleAssembler::emit_i_times(const ArgLabel &Fail,
-                                       const ArgSource &LHS,
-                                       const ArgSource &RHS,
-                                       const ArgRegister &Dst) {
-    bool small_result = is_product_small_if_args_are_small(LHS, RHS);
+/* ARG2 = Src1
+ * ARG3 = Src2
+ *
+ * The result is returned in RET.
+ */
+void BeamGlobalAssembler::emit_mul_body_shared() {
+    mov_imm(ARG4, make_small(0));
+    a.jmp(labels[mul_add_body_shared]);
+}
 
-    if (always_small(LHS) && always_small(RHS) && small_result) {
-        comment("multiplication without overflow check");
-        if (RHS.isSmall()) {
-            Sint factor = RHS.as<ArgSmall>().getSigned();
+/* ARG2 = Src1
+ * ARG3 = Src2
+ *
+ * Result is returned in RET, error is indicated by ZF.
+ */
+void BeamGlobalAssembler::emit_mul_guard_shared() {
+    mov_imm(ARG4, make_small(0));
+    a.jmp(labels[mul_add_guard_shared]);
+}
 
-            mov_arg(RET, LHS);
+void BeamModuleAssembler::emit_i_mul_add(const ArgLabel &Fail,
+                                         const ArgSource &Src1,
+                                         const ArgSource &Src2,
+                                         const ArgSource &Src3,
+                                         const ArgSource &Src4,
+                                         const ArgRegister &Dst) {
+    bool is_product_small = is_product_small_if_args_are_small(Src1, Src2);
+    bool is_sum_small = is_sum_small_if_args_are_small(Src3, Src4);
+    bool sometimes_small = !(Src2.isLiteral() || Src4.isLiteral());
+    bool is_increment_zero =
+            Src4.isSmall() && Src4.as<ArgSmall>().getSigned() == 0;
+    Sint factor = 0;
+    int left_shift = -1;
+
+    if (is_increment_zero) {
+        comment("(adding zero)");
+    }
+
+    if (Src2.isSmall()) {
+        factor = Src2.as<ArgSmall>().getSigned();
+        if (Support::isPowerOf2(factor)) {
+            left_shift = Support::ctz<Eterm>(factor);
+        }
+    }
+
+    if (always_small(Src1) && Src2.isSmall() && Src4.isSmall() &&
+        is_product_small && is_sum_small) {
+        x86::Mem p;
+        Sint increment = Src4.as<ArgSmall>().get();
+        increment -= factor * _TAG_IMMED1_SMALL;
+
+        switch (factor) {
+        case 2:
+            p = ptr(RET, RET, 0, increment);
+            break;
+        case 3:
+            p = ptr(RET, RET, 1, increment);
+            break;
+        case 4:
+            p = ptr(x86::Gp(), RET, 2, increment);
+            break;
+        case 5:
+            p = ptr(RET, RET, 2, increment);
+            break;
+        case 8:
+            p = ptr(x86::Gp(), RET, 3, increment);
+            break;
+        case 9:
+            p = ptr(RET, RET, 3, increment);
+            break;
+        }
+
+        if (Support::isInt32(increment) && p.hasIndex()) {
+            comment("optimizing multiplication and addition using LEA");
+            mov_arg(RET, Src1);
+            a.lea(RET, p);
+            mov_arg(Dst, RET);
+            return;
+        }
+    }
+
+    if (always_small(Src1) && always_small(Src2) && always_small(Src4) &&
+        is_product_small && is_sum_small) {
+        comment("multiplication and addition without overflow check");
+        if (Src2.isSmall()) {
+            mov_arg(RET, Src1);
             a.and_(RET, imm(~_TAG_IMMED1_MASK));
             if (Support::isPowerOf2(factor)) {
-                int trailing_bits = Support::ctz<Eterm>(factor);
                 comment("optimized multiplication by replacing with left "
                         "shift");
-                a.shl(RET, imm(trailing_bits));
+                a.shl(RET, imm(left_shift));
             } else {
                 mov_imm(ARG2, factor);
                 a.imul(RET, ARG2);
             }
         } else {
-            mov_arg(RET, LHS);
-            mov_arg(ARG2, RHS);
+            mov_arg(RET, Src1);
+            mov_arg(ARG2, Src2);
             a.and_(RET, imm(~_TAG_IMMED1_MASK));
             a.sar(ARG2, imm(_TAG_IMMED1_SIZE));
             a.imul(RET, ARG2);
         }
 
-        a.or_(RET, imm(_TAG_IMMED1_SMALL));
+        if (is_increment_zero) {
+            a.or_(RET, imm(_TAG_IMMED1_SMALL));
+        } else {
+            mov_arg(ARG2, Src4);
+            a.add(RET, ARG2);
+        }
         mov_arg(Dst, RET);
 
         return;
@@ -917,39 +1125,85 @@ void BeamModuleAssembler::emit_i_times(const ArgLabel &Fail,
 
     Label next = a.newLabel(), mixed = a.newLabel();
 
-    mov_arg(ARG2, LHS); /* Used by erts_mixed_times in this slot */
-    mov_arg(ARG3, RHS); /* Used by erts_mixed_times in this slot */
-
-    if (RHS.isSmall()) {
-        Sint val = RHS.as<ArgSmall>().getSigned();
-        emit_is_small(mixed, LHS, ARG2);
-        a.mov(RET, ARG2);
-        a.mov(ARG4, imm(val));
-    } else {
-        emit_are_both_small(mixed, LHS, ARG2, RHS, ARG3);
-        a.mov(RET, ARG2);
-        a.mov(ARG4, ARG3);
-        a.sar(ARG4, imm(_TAG_IMMED1_SIZE));
+    mov_arg(ARG2, Src1);
+    mov_arg(ARG3, Src2);
+    if (!is_increment_zero) {
+        mov_arg(ARG4, Src4);
     }
 
-    a.and_(RET, imm(~_TAG_IMMED1_MASK));
-    a.imul(RET, ARG4);
-    if (small_result) {
-        comment("skipped overflow check because the result is always small");
+    if (!sometimes_small) {
+        comment("skipped test for small because one operand is never small");
+    } else if (Src2.isSmall()) {
+        Sint val = Src2.as<ArgSmall>().getSigned();
+        emit_are_both_small(mixed, Src1, ARG2, Src4, ARG4);
+        a.mov(RET, ARG2);
+        mov_imm(ARG5, val);
     } else {
-        a.short_().jo(mixed);
+        if (is_increment_zero) {
+            emit_are_both_small(mixed, Src1, ARG2, Src2, ARG3);
+        } else if (always_small(Src1)) {
+            emit_are_both_small(mixed, Src2, ARG3, Src4, ARG4);
+        } else {
+            a.mov(RETd, ARG2.r32());
+            a.and_(RETd, ARG3.r32());
+            a.and_(RETd, ARG4.r32());
+            if (always_one_of<BeamTypeId::Integer, BeamTypeId::AlwaysBoxed>(
+                        Src1) &&
+                always_one_of<BeamTypeId::Integer, BeamTypeId::AlwaysBoxed>(
+                        Src2) &&
+                always_one_of<BeamTypeId::Integer, BeamTypeId::AlwaysBoxed>(
+                        Src4)) {
+                emit_is_not_boxed(mixed, RET);
+            } else {
+                a.and_(RETb, imm(_TAG_IMMED1_MASK));
+                a.cmp(RETb, imm(_TAG_IMMED1_SMALL));
+                a.short_().jne(mixed);
+            }
+        }
+        a.mov(RET, ARG2);
+        a.mov(ARG5, ARG3);
+        a.sar(ARG5, imm(_TAG_IMMED1_SIZE));
+    }
+
+    if (sometimes_small) {
+        a.and_(RET, imm(~_TAG_IMMED1_MASK));
+        a.imul(RET, ARG5);
+        if (is_product_small) {
+            comment("skipped overflow check because product is always small");
+        } else {
+            a.short_().jo(mixed);
+        }
+
+        if (is_increment_zero) {
+            a.or_(RET, imm(_TAG_IMMED1_SMALL));
+        } else {
+            a.add(RET, ARG4);
+            if (is_sum_small) {
+                comment("skipped overflow check because sum is always small");
+            } else {
+                a.short_().jo(mixed);
+            }
+        }
+
+        a.short_().jmp(next);
     }
-    a.or_(RET, imm(_TAG_IMMED1_SMALL));
-    a.short_().jmp(next);
 
     /* Call mixed multiplication. */
     a.bind(mixed);
     {
         if (Fail.get() != 0) {
-            safe_fragment_call(ga->get_times_guard_shared());
+            if (is_increment_zero) {
+                safe_fragment_call(ga->get_mul_guard_shared());
+            } else {
+                safe_fragment_call(ga->get_mul_add_guard_shared());
+            }
             a.je(resolve_beam_label(Fail));
         } else {
-            safe_fragment_call(ga->get_times_body_shared());
+            if (is_increment_zero) {
+                safe_fragment_call(ga->get_mul_body_shared());
+            } else {
+                safe_fragment_call(ga->get_mul_add_body_shared());
+            }
         }
     }
 
@@ -960,15 +1214,15 @@ void BeamModuleAssembler::emit_i_times(const ArgLabel &Fail,
 /* ARG2 (!) = LHS, RET (!) = RHS
  *
  * Result is returned in RET. Error is indicated by ZF. */
-template<typename T>
-void BeamGlobalAssembler::emit_bitwise_fallback_guard(T(*func_ptr)) {
+template<typename T, T Func>
+void BeamGlobalAssembler::emit_bitwise_fallback_guard() {
     emit_enter_frame();
     emit_enter_runtime();
 
     a.mov(ARG1, c_p);
     /* ARG2 is already set to LHS */
     a.mov(ARG3, RET);
-    runtime_call<3>(func_ptr);
+    runtime_call<T, Func>();
 
     emit_leave_runtime();
     emit_leave_frame();
@@ -980,9 +1234,8 @@ void BeamGlobalAssembler::emit_bitwise_fallback_guard(T(*func_ptr)) {
 /* ARG2 (!) = LHS, RET (!) = RHS
  *
  * Result is returned in RET. */
-template<typename T>
-void BeamGlobalAssembler::emit_bitwise_fallback_body(T(*func_ptr),
-                                                     const ErtsCodeMFA *mfa) {
+template<typename T, T Func>
+void BeamGlobalAssembler::emit_bitwise_fallback_body(const ErtsCodeMFA *mfa) {
     Label error = a.newLabel();
 
     emit_enter_frame();
@@ -995,7 +1248,7 @@ void BeamGlobalAssembler::emit_bitwise_fallback_body(T(*func_ptr),
     a.mov(ARG1, c_p);
     /* ARG2 is already set to LHS */
     a.mov(ARG3, RET);
-    runtime_call<3>(func_ptr);
+    runtime_call<T, Func>();
 
     emit_leave_runtime();
     emit_leave_frame();
@@ -1018,12 +1271,14 @@ void BeamGlobalAssembler::emit_bitwise_fallback_body(T(*func_ptr),
 }
 
 void BeamGlobalAssembler::emit_i_band_guard_shared() {
-    emit_bitwise_fallback_guard(erts_band);
+    emit_bitwise_fallback_guard<Eterm (*)(Process *, Eterm, Eterm),
+                                erts_band>();
 }
 
 void BeamGlobalAssembler::emit_i_band_body_shared() {
     static const ErtsCodeMFA bif_mfa = {am_erlang, am_band, 2};
-    emit_bitwise_fallback_body(erts_band, &bif_mfa);
+    emit_bitwise_fallback_body<Eterm (*)(Process *, Eterm, Eterm), erts_band>(
+            &bif_mfa);
 }
 
 void BeamModuleAssembler::emit_i_band(const ArgSource &LHS,
@@ -1034,10 +1289,18 @@ void BeamModuleAssembler::emit_i_band(const ArgSource &LHS,
         comment("skipped test for small operands since they are always small");
         mov_arg(RET, LHS);
         if (RHS.isSmall() && Support::isInt32(RHS.as<ArgSmall>().get())) {
-            a.and_(RETd, imm(RHS.as<ArgSmall>().get()));
+            preserve_cache(
+                    [&]() {
+                        a.and_(RETd, imm(RHS.as<ArgSmall>().get()));
+                    },
+                    RET);
         } else if (RHS.isSmall() &&
                    Support::isInt32((Sint)RHS.as<ArgSmall>().get())) {
-            a.and_(RET, imm(RHS.as<ArgSmall>().get()));
+            preserve_cache(
+                    [&]() {
+                        a.and_(RET, imm(RHS.as<ArgSmall>().get()));
+                    },
+                    RET);
         } else {
             mov_arg(ARG2, RHS);
             a.and_(RET, ARG2);
@@ -1051,11 +1314,15 @@ void BeamModuleAssembler::emit_i_band(const ArgSource &LHS,
 
     Label generic = a.newLabel(), next = a.newLabel();
 
-    emit_are_both_small(generic, LHS, ARG2, RHS, RET);
+    if (RHS.isLiteral()) {
+        comment("skipped test for small because one operand is never small");
+    } else {
+        emit_are_both_small(generic, LHS, ARG2, RHS, RET);
 
-    /* TAG & TAG = TAG, so we don't need to tag it again. */
-    a.and_(RET, ARG2);
-    a.short_().jmp(next);
+        /* TAG & TAG = TAG, so we don't need to tag it again. */
+        a.and_(RET, ARG2);
+        a.short_().jmp(next);
+    }
 
     a.bind(generic);
     {
@@ -1075,7 +1342,7 @@ void BeamModuleAssembler::emit_i_band(const ArgSource &LHS,
  *
  * Result is returned in RET. Error is indicated by ZF. */
 void BeamGlobalAssembler::emit_i_bor_guard_shared() {
-    emit_bitwise_fallback_guard(erts_bor);
+    emit_bitwise_fallback_guard<Eterm (*)(Process *, Eterm, Eterm), erts_bor>();
 }
 
 /* ARG2 (!) = LHS, RET (!) = RHS
@@ -1083,7 +1350,8 @@ void BeamGlobalAssembler::emit_i_bor_guard_shared() {
  * Result is returned in RET. */
 void BeamGlobalAssembler::emit_i_bor_body_shared() {
     static const ErtsCodeMFA bif_mfa = {am_erlang, am_bor, 2};
-    emit_bitwise_fallback_body(erts_bor, &bif_mfa);
+    emit_bitwise_fallback_body<Eterm (*)(Process *, Eterm, Eterm), erts_bor>(
+            &bif_mfa);
 }
 
 void BeamModuleAssembler::emit_i_bor(const ArgLabel &Fail,
@@ -1132,7 +1400,8 @@ void BeamModuleAssembler::emit_i_bor(const ArgLabel &Fail,
  *
  * Result is returned in RET. Error is indicated by ZF. */
 void BeamGlobalAssembler::emit_i_bxor_guard_shared() {
-    emit_bitwise_fallback_guard(erts_bxor);
+    emit_bitwise_fallback_guard<Eterm (*)(Process *, Eterm, Eterm),
+                                erts_bxor>();
 }
 
 /* ARG2 (!) = LHS, RET (!) = RHS
@@ -1140,7 +1409,8 @@ void BeamGlobalAssembler::emit_i_bxor_guard_shared() {
  * Result is returned in RET. */
 void BeamGlobalAssembler::emit_i_bxor_body_shared() {
     static const ErtsCodeMFA bif_mfa = {am_erlang, am_bxor, 2};
-    emit_bitwise_fallback_body(erts_bxor, &bif_mfa);
+    emit_bitwise_fallback_body<Eterm (*)(Process *, Eterm, Eterm), erts_bxor>(
+            &bif_mfa);
 }
 
 void BeamModuleAssembler::emit_i_bxor(const ArgLabel &Fail,
@@ -1151,12 +1421,22 @@ void BeamModuleAssembler::emit_i_bxor(const ArgLabel &Fail,
         comment("skipped test for small operands since they are always small");
         mov_arg(RET, LHS);
         if (RHS.isImmed() && Support::isInt32((Sint)RHS.as<ArgSmall>().get())) {
-            a.xor_(RET, imm(RHS.as<ArgSmall>().get() & ~_TAG_IMMED1_SMALL));
+            preserve_cache(
+                    [&]() {
+                        a.xor_(RET,
+                               imm(RHS.as<ArgSmall>().get() &
+                                   ~_TAG_IMMED1_SMALL));
+                    },
+                    RET);
         } else {
-            /* TAG ^ TAG = 0, so we need to tag it again. */
             mov_arg(ARG2, RHS);
-            a.xor_(RET, ARG2);
-            a.or_(RET, imm(_TAG_IMMED1_SMALL));
+            preserve_cache(
+                    [&]() {
+                        /* TAG ^ TAG = 0, so we need to tag it again. */
+                        a.xor_(RET, ARG2);
+                        a.or_(RET, imm(_TAG_IMMED1_SMALL));
+                    },
+                    RET);
         }
         mov_arg(Dst, RET);
         return;
@@ -1201,7 +1481,7 @@ void BeamGlobalAssembler::emit_i_bnot_guard_shared() {
 
     a.mov(ARG1, c_p);
     a.mov(ARG2, RET);
-    runtime_call<2>(erts_bnot);
+    runtime_call<Eterm (*)(Process *, Eterm), erts_bnot>();
 
     emit_leave_runtime();
     emit_leave_frame();
@@ -1230,7 +1510,7 @@ void BeamGlobalAssembler::emit_i_bnot_body_shared() {
 
     a.mov(ARG1, c_p);
     a.mov(ARG2, RET);
-    runtime_call<2>(erts_bnot);
+    runtime_call<Eterm (*)(Process *, Eterm), erts_bnot>();
 
     emit_leave_runtime();
     emit_leave_frame();
@@ -1288,7 +1568,7 @@ void BeamModuleAssembler::emit_i_bnot(const ArgLabel &Fail,
  *
  * Result is returned in RET. Error is indicated by ZF. */
 void BeamGlobalAssembler::emit_i_bsr_guard_shared() {
-    emit_bitwise_fallback_guard(erts_bsr);
+    emit_bitwise_fallback_guard<Eterm (*)(Process *, Eterm, Eterm), erts_bsr>();
 }
 
 /* ARG2 (!) = LHS, RET (!) = RHS
@@ -1296,7 +1576,8 @@ void BeamGlobalAssembler::emit_i_bsr_guard_shared() {
  * Result is returned in RET. */
 void BeamGlobalAssembler::emit_i_bsr_body_shared() {
     static const ErtsCodeMFA bif_mfa = {am_erlang, am_bsr, 2};
-    emit_bitwise_fallback_body(erts_bsr, &bif_mfa);
+    emit_bitwise_fallback_body<Eterm (*)(Process *, Eterm, Eterm), erts_bsr>(
+            &bif_mfa);
 }
 
 void BeamModuleAssembler::emit_i_bsr(const ArgSource &LHS,
@@ -1305,13 +1586,14 @@ void BeamModuleAssembler::emit_i_bsr(const ArgSource &LHS,
                                      const ArgRegister &Dst) {
     Label generic = a.newLabel(), next = a.newLabel();
     bool need_generic = true;
+    bool need_register_load = true;
 
     mov_arg(ARG2, LHS);
 
     if (RHS.isSmall()) {
         Sint shift = RHS.as<ArgSmall>().getSigned();
 
-        if (shift >= 0 && shift < SMALL_BITS - 1) {
+        if (shift >= 0) {
             if (always_small(LHS)) {
                 comment("skipped test for small left operand because it is "
                         "always small");
@@ -1325,6 +1607,7 @@ void BeamModuleAssembler::emit_i_bsr(const ArgSource &LHS,
             /* We don't need to clear the mask after shifting because
              * _TAG_IMMED1_SMALL will set all the bits anyway. */
             ERTS_CT_ASSERT(_TAG_IMMED1_MASK == _TAG_IMMED1_SMALL);
+            shift = std::min<Sint>(shift, 63);
             a.sar(RET, imm(shift));
             a.or_(RET, imm(_TAG_IMMED1_SMALL));
 
@@ -1332,14 +1615,33 @@ void BeamModuleAssembler::emit_i_bsr(const ArgSource &LHS,
                 a.short_().jmp(next);
             }
         } else {
-            /* Constant shift is negative or too big to fit the `sar`
-             * instruction, fall back to the generic path. */
+            /* Constant shift is negative; fall back to the generic
+             * path. */
         }
+    } else if (hasCpuFeature(CpuFeatures::X86::kBMI2)) {
+        mov_arg(RET, RHS);
+        need_register_load = false;
+
+        emit_are_both_small(generic, LHS, ARG2, RHS, RET);
+
+        a.mov(ARG1, RET);
+        a.sar(ARG1, imm(_TAG_IMMED1_SIZE));
+        a.js(generic);
+
+        mov_imm(RET, 63);
+        a.cmp(ARG1, RET);
+        a.cmova(ARG1, RET);
+
+        a.sarx(RET, ARG2, ARG1);
+        a.or_(RET, imm(_TAG_IMMED1_SMALL));
+        a.short_().jmp(next);
     }
 
     a.bind(generic);
     if (need_generic) {
-        mov_arg(RET, RHS);
+        if (need_register_load) {
+            mov_arg(RET, RHS);
+        }
 
         if (Fail.get() != 0) {
             safe_fragment_call(ga->get_i_bsr_guard_shared());
@@ -1357,7 +1659,7 @@ void BeamModuleAssembler::emit_i_bsr(const ArgSource &LHS,
  *
  * Result is returned in RET. Error is indicated by ZF. */
 void BeamGlobalAssembler::emit_i_bsl_guard_shared() {
-    emit_bitwise_fallback_guard(erts_bsl);
+    emit_bitwise_fallback_guard<Eterm (*)(Process *, Eterm, Eterm), erts_bsl>();
 }
 
 /* ARG2 (!) = LHS, RET (!) = RHS
@@ -1365,7 +1667,8 @@ void BeamGlobalAssembler::emit_i_bsl_guard_shared() {
  * Result is returned in RET. */
 void BeamGlobalAssembler::emit_i_bsl_body_shared() {
     static const ErtsCodeMFA bif_mfa = {am_erlang, am_bsl, 2};
-    emit_bitwise_fallback_body(erts_bsl, &bif_mfa);
+    emit_bitwise_fallback_body<Eterm (*)(Process *, Eterm, Eterm), erts_bsl>(
+            &bif_mfa);
 }
 
 static int count_leading_zeroes(UWord value) {
