@@ -87,6 +87,13 @@ BOOLEAN_T esock_get_bool_from_map(ErlNifEnv*   env,
                                   ERL_NIF_TERM key,
                                   BOOLEAN_T    def);
 
+extern
+BOOLEAN_T esock_get_string_from_map(ErlNifEnv*         env,
+                                    ERL_NIF_TERM       map,
+                                    ERL_NIF_TERM       key,
+                                    ErlNifCharEncoding encoding,
+                                    char**             str);
+
 extern
 BOOLEAN_T esock_decode_iov(ErlNifEnv*    env,
                            ERL_NIF_TERM  eIOV,
@@ -320,7 +327,15 @@ BOOLEAN_T esock_timestamp_str(char *buf, unsigned int len);
 extern
 BOOLEAN_T esock_format_timestamp(ErlNifTime timestamp, char *buf, unsigned int len);
 
+/*
+ * esock_debug_msg
+ * esock_info_msg
+ * esock_warning_msg
+ * esock_error_msg
+ */
+
 #define MSG_FUNCS                               \
+    MSG_FUNC_DEF(debug)                         \
     MSG_FUNC_DEF(info)                          \
     MSG_FUNC_DEF(warning)                       \
     MSG_FUNC_DEF(error)
