@@ -1,6 +1,8 @@
 %%
 %% %CopyrightBegin%
 %%
+%% SPDX-License-Identifier: Apache-2.0
+%%
 %% Copyright Ericsson AB 1997-2025. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
@@ -102,8 +104,6 @@ bit_string(Rules, Opts) ->
 	{ber,[]} ->
 	    %% XXX
 	    ok;
-        {jer,[]} ->
-            ok;
 	{_,_} ->
 	    %% DER, PER, UPER
 	    consistent_def_enc('BsDef1',
@@ -238,7 +238,6 @@ bit_string(Rules, Opts) ->
 
     case Rules of
 	ber -> ok;
-        jer -> ok;
 	_ -> per_bs_strings()
     end.
 
