@@ -1,7 +1,9 @@
 %%
 %% %CopyrightBegin%
+%%
+%% SPDX-License-Identifier: Apache-2.0
 %% 
-%% Copyright Ericsson AB 2012-2023. All Rights Reserved.
+%% Copyright Ericsson AB 2012-2025. All Rights Reserved.
 %% 
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -485,8 +487,8 @@ app_start(App, Config) ->
     try
 	case App of
 	    crypto ->
-		crypto:stop(),
-		ok = crypto:start();
+		application:stop(crypto),
+		ok = application:start(crypto);
 	    inets ->
 		application:stop(App),
 		ok = application:start(App),
