@@ -1,8 +1,10 @@
 %%
 %% %CopyrightBegin%
-%% 
-%% Copyright Ericsson AB 2018-2023. All Rights Reserved.
-%% 
+%%
+%% SPDX-License-Identifier: Apache-2.0
+%%
+%% Copyright Ericsson AB 2018-2025. All Rights Reserved.
+%%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
 %% You may obtain a copy of the License at
@@ -14,7 +16,7 @@
 %% WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 %% See the License for the specific language governing permissions and
 %% limitations under the License.
-%% 
+%%
 %% %CopyrightEnd%
 %%
 
@@ -619,9 +621,24 @@ reader_loop(#{active      := once,
                     Pid ! ?CLOSED_MSG(Sock, Method),
                     reader_exit(State, E1);
 
-                {error, Reason} = E2 ->
+                {error, {Reason, _}} ->
+                    E2 = {error, Reason},
                     Pid ! ?ERROR_MSG(Sock, Method, Reason),
-                    reader_exit(State, E2)
+                    reader_exit(State, E2);
+
+                {error, Reason} = E3 ->
+                    Pid ! ?ERROR_MSG(Sock, Method, Reason),
+                    reader_exit(State, E3);
+
+                {select, SelectInfo} ->
+                    %% Why would this happen?
+                    %% We get a select message telling us that
+                    %% something happened on the socket (data or maybe
+                    %% an error) but then when we try to read
+                    %% we get nothing...
+                    reader_loop(State#{asynch_info => SelectInfo,
+                                       asynch_num  => ANum+1})
+
             end;
 
         ?COMPLETION_MSG(Sock, Ref, Result) ->
@@ -1349,12 +1366,12 @@ reader_exit(#{async       := true,
 
 %% ==========================================================================
 
-mq() ->
-    mq(self()).
+%% mq() ->
+%%     mq(self()).
 
-mq(Pid) when is_pid(Pid) ->
-    {messages, MQ} = process_info(Pid, messages),
-    MQ.
+%% mq(Pid) when is_pid(Pid) ->
+%%     {messages, MQ} = process_info(Pid, messages),
+%%     MQ.
 
              
 %% ==========================================================================
