@@ -1,7 +1,9 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 2007-2024. All Rights Reserved.
+%% SPDX-License-Identifier: Apache-2.0
+%%
+%% Copyright Ericsson AB 2007-2025. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -67,7 +69,8 @@
          groups/1,
          group_to_enum/1,
          enum_to_group/1,
-         default_groups/0]).
+         default_groups/0,
+         default_suites/1]).
 
 -export([derive_secret/4,
          hkdf_expand_label/5,
