@@ -1,7 +1,9 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 2010-2024. All Rights Reserved.
+%% SPDX-License-Identifier: Apache-2.0
+%%
+%% Copyright Ericsson AB 2010-2025. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -85,16 +87,17 @@ normal(Config) when is_list(Config) ->
     {yes,"test:",[]} = do_expand("expand_"),
     {no, [], []} = do_expand("expandXX_"),
     {no,[],[#{
-               title:="functions",
+               title:="Functions",
                elems:=[{"a_fun_name",[{ending,"("}]},
                         {"a_less_fun_name",_},
                         {"b_comes_after_a",_},
                         {"expand0arity_entirely",_},
                         {"module_info",_}]
               }]} = do_expand("expand_test:"),
-    {yes,[],[#{title:="functions",
+    {yes,[],[#{title:="Functions",
                       elems:=[{"a_fun_name",_},{"a_less_fun_name",_}]}]} = do_expand("expand_test:a_"),
     {yes,"arity_entirely()",[]} = do_expand("expand_test:expand0"),
+    {no, [], [#{title:="Functions"}, #{title:="Obsolete API functions"}]} = do_expand("string:"),
     ok.
 
 to_atom(Str) ->
@@ -109,13 +112,14 @@ type_completion(_Config) ->
     ct:timetrap({minutes, 20}),
     {Time,_} = timer:tc(fun() -> do_expand("erl_pp:expr(") end),
     case Time of
-        Time when Time > 2600000 -> {skip, "Expansion too slow on this machine"};
+        Time when Time > 5000000 ->
+            {skip, lists:flatten(io_lib:format("Expansion too slow (~p) on this machine",[Time]))};
         _ ->
             parallelforeach(
             fun(Mod) ->
                     Exports = edlin_expand:get_exports(Mod),
                     [try
-                        Str = io_lib:write_atom(Mod) ++ ":" ++ atom_to_list(Func) ++ "(",
+                        Str = io_lib:write_atom(Mod) ++ ":" ++ io_lib:write_atom(Func) ++ "(",
                         do_expand(Str)
                     catch E:R:ST ->
                             erlang:raise(E, {R, Mod, Func}, ST)
@@ -258,6 +262,10 @@ function_parameter_completion(Config) ->
     {no, [], [#{elems:=[#{elems:=[#{elems:=[{"any()",[]},{"[any() | [Deeplist]]",[]}]}]}]}]} = do_expand("complete_function_parameter:a_deeplist_fun("),
     {no,[],[#{title:="typespecs",
                 elems:=[#{title:=
+                                "complete_function_parameter:multi_arity_fun()",
+                            options:=[],
+                            elems:=[{")",[]}]},
+                        #{title:=
                                "complete_function_parameter:multi_arity_fun(T1)",
                           elems:=[#{title:="types",
                                     elems:=[{"integer()",[]}],
@@ -268,13 +276,10 @@ function_parameter_completion(Config) ->
                             elems:=[#{title:="types",
                                     elems:=[{"integer()",[]}],
                                     options:=[{hide,title}]}],
-                            options:=[{highlight_param,1}]},
-                            #{title:=
-                                "complete_function_parameter:multi_arity_fun()",
-                            options:=[],
-                            elems:=[")"]}],
+                            options:=[{highlight_param,1}]}],
                 options:=[highlight_all]}]} = do_expand("complete_function_parameter:multi_arity_fun("),
     {no, [], [#{elems:=[#{elems:=[#{elems:=[{"true",[]},{"false",[]}]}]}]}]} = do_expand("complete_function_parameter:multi_arity_fun(1,"),
+    {no, [], []} = do_expand("complete_function_parameter:multi_arity_fun(["),
     {no,[],
         [#{elems :=
             [#{elems :=
@@ -361,6 +366,10 @@ get_coverage(Config) ->
     do_expand("complete_function_parameter:map_parameter_function(#{}, "),
     do_expand("complete_function_parameter:map_parameter_function(#{V=>1}, "),
     do_expand("complete_function_parameter:map_parameter_function(#{a=>V}, "),
+    do_expand("complete_function_parameter:map_variable_parameter_function(#{"),
+    do_expand("complete_function_parameter:map_variable_parameter_function(#{a"),
+    do_expand("complete_function_parameter:map_variable_parameter_function(#{a => "),
+    do_expand("complete_function_parameter:map_variable_parameter_function(#{a => a"),
     do_expand("complete_function_parameter:tuple_parameter_function({a,b}, "),
     do_expand("complete_function_parameter:tuple_parameter_function({a,V}, "),
     do_expand("complete_function_parameter:list_parameter_function([], "),
@@ -442,8 +451,11 @@ get_coverage(Config) ->
     do_expand("M#"),
     do_expand("#non_existant_record"),
     do_expand("#a_record{ non_existand_field"),
-    
-    
+    do_expand("case("),
+    do_expand("catch("),
+    do_expand("case ("),
+    do_expand("catch ("),
+
     %% match_arguments coverage
     do_expand("complete_function_parameter:integer_parameter_function(atom,"), %% match_argument -> false
     do_expand("complete_function_parameter:a_zero_arity_fun()"), %% match_argument, parameters empty
@@ -518,7 +530,7 @@ get_coverage(Config) ->
     lists:flatten(edlin_expand:format_matches(M10, 20)),
     %% Test that we are not filtering duplicates bit with different case or different string lengths
     {yes,"e", M11} = do_expand("complete_function_parameter:cas"),
-    "\e[;1;4mfunctions\e[0m\ncaseSensitiveFunction(        casesensitivefunction(        \ncaseSensitiveFunctionName(\n" = do_format(M11),
+    "\e[;1;4mFunctions\e[0m\ncaseSensitiveFunction(        casesensitivefunction(        \ncaseSensitiveFunctionName(\n" = do_format(M11),
     ok.
 
 %% Normal module name, some function names using quoted atoms.
@@ -528,7 +540,7 @@ quoted_fun(Config) when is_list(Config) ->
     %% should be no colon after test this time
     {yes, "test", [#{title:="modules", elems:=[{"expand_test",[{ending, ":"}]},{"expand_test1",_}]}]} = do_expand("expand_"),
     {no, [], []} = do_expand("expandXX_"),
-    {no,[],[#{title:="functions",
+    {no,[],[#{title:="Functions",
                      elems:=[{"'#weird-fun-name'",_},
                             {"'Quoted_fun_name'",_},
                             {"'Quoted_fun_too'",_},
@@ -561,7 +573,7 @@ quoted_module(Config) when is_list(Config) ->
                             {"a_less_fun_name",_},
                             {"b_comes_after_a",_},
                             {"module_info",_}]}]} = do_expand("'ExpandTestCaps':"),
-    {yes,[],[#{title:="functions", elems:=[{"a_fun_name",_},
+    {yes,[],[#{title:="Functions", elems:=[{"a_fun_name",_},
                                                 {"a_less_fun_name",_}]}]} = do_expand("'ExpandTestCaps':a_"),
     ok.
 
@@ -634,7 +646,7 @@ unicode(Config) when is_list(Config) ->
     ok.
 
 do_expand(String) ->
-    % erlang:display(String),
+    io:format("~ts", [String]),
     Bs = [
           {'Binding', 0},
           {'MapBinding', #{a_key=>0, b_key=>1, c_key=>2}},
